'''
Function:
    Seach and download papers from arxiv
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
from .base import Base
from bs4 import BeautifulSoup


'''Seach and download papers from arxiv'''
class Arxiv(Base):
    def __init__(self, config=None, logger_handle=None, **kwargs):
        super(Arxiv, self).__init__(config, logger_handle, **kwargs)
        self.source = 'arxiv'
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36'
        }
    '''search paper'''
    def search(self, keyword):
        # search
        keyword = keyword.replace(' ', '+')
        search_url = 'https://arxiv.org/search/?'
        params = {
            'query': keyword,
            'searchtype': 'all',
            'abstracts': 'show',
            'order': '-announced_date_first',
            'size': '50',
        }
        response = self.session.get(search_url, params=params, headers=self.headers)
        # parse
        soup = BeautifulSoup(response.text, features='lxml')
        paperinfos = []
        for item in soup.find('ol').find_all('li', attrs={'class': 'arxiv-result'}):
            try: title = item.find('p', attrs={'class': 'title'}).text.strip()
            except: title = ''
            try: authors = self.cleantext(item.find('p', attrs={'class': 'authors'}).text.strip()).replace('Authors:', '')
            except: authors = ''
            try: url = item.find('p', attrs={'class': 'list-title'}).find('a').attrs['href']
            except: url = ''
            paperinfo = {
                'source': self.source,
                'savedir': self.config['savedir'],
                'ext': 'pdf',
                'savename': title,
                'title': title,
                'authors': authors,
                'download_url': url.replace('abs', 'pdf') + '.pdf',
            }
            paperinfos.append(paperinfo)
            if len(paperinfos) == self.config['search_size_per_source']: break
        # return
        return paperinfos
    '''clean text'''
    def cleantext(self, text):
        text = text.replace('\n', '')
        text, text_clean = text.split(' '), []
        for item in text:
            if item: text_clean.append(item)
        text = ' '.join(text_clean)
        return text


'''
Function:
    Seach and download papers from Sci-hub
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
import re
from .base import Base
from lxml import etree
from ..utils import Downloader
from urllib.parse import urlparse


'''Seach and download papers from Sci-hub'''
class SciHub(Base):
    def __init__(self, config=None, logger_handle=None, **kwargs):
        super(SciHub, self).__init__(config, logger_handle, **kwargs)
        self.source = 'scihub'
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36'
        }
    '''parse paper infos before dowload paper'''
    def parseinfosbeforedownload(self, paperinfos):
        scihub_sites = [
            'https://sci-hub.st/', 'https://sci-hub.ru/', 'https://sci-hub.se/', 
            'https://sci-hub.shop/', 'http://sci-hub.ren/', 
        ]
        # fetch pdf url
        for paperinfo in paperinfos:
            input_content = paperinfo['input']
            input_type = self.guessinputtype(input_content)
            if input_type == 'pdf': 
                paperinfo['download_url'] = input_content
            elif input_type == 'arxiv':
                paperinfo['download_url'] = input_content.replace('abs', 'pdf') + '.pdf'
            else:
                for scihub_site in scihub_sites:
                    self.headers.update({'Referer': scihub_site})
                    try: response = self.session.get(scihub_site + input_content, headers=self.headers, verify=False)
                    except: continue
                    if response.status_code not in [200]: continue
                    html = etree.HTML(response.content)
                    article = html.xpath('//div[@id="article"]/embed[1]') or html.xpath('//div[@id="article"]/iframe[1]') if html is not None else None
                    if not article: continue
                    pdf_url = urlparse(article[0].attrib['src'], scheme='http').geturl()
                    paperinfo['download_url'] = pdf_url
                    break
            if 'download_url' not in paperinfo: paperinfo['download_url'] = None
            paperinfo['source'] = self.source
        # return
        return paperinfos
    '''guess input type'''
    def guessinputtype(self, input_content):
        input_type, doi_pattern = None, re.compile(r'\b(10[.][0-9]{4,}(?:[.][0-9]+)*/(?:(?!["&\'])\S)+)\b')
        if input_content.startswith('http') or input_content.startswith('https'):
            if '.pdf' in input_content: input_type = 'pdf'
            elif 'arxiv' in input_content: input_type = 'arxiv'
            else: input_type = 'url'
        elif input_content.isdigit(): input_type = 'pmid'
        elif input_content.startswith('doi:') or doi_pattern.match(input_content): input_type = 'doi'
        else: input_type = 'string'
        return input_type


'''
Function:
    Seach and download papers from google scholar
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
from .scihub import SciHub
from bs4 import BeautifulSoup


'''Seach and download papers from google scholar'''
class GoogleScholar(SciHub):
    def __init__(self, config=None, logger_handle=None, **kwargs):
        super(GoogleScholar, self).__init__(config, logger_handle, **kwargs)
        self.source = 'googlescholar'
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36'
        }
    '''search paper'''
    def search(self, keyword):
        # search
        if self.config.get('area') == 'CN':
            search_urls = [
                'https://xs.dailyheadlines.cc/scholar',
                'https://xs2.dailyheadlines.cc/scholar',
                'https://scholar.lanfanshu.cn/scholar',
            ]
            params = {
                'hl': 'zh-CN',
                'as_sdt': '0,33',
                'q': keyword,
                'btnG': ''
            }
            for search_url in search_urls:
                self.headers.update({'Referer': f'{search_url}?q={keyword}'})
                response = self.session.get(search_url, params=params, headers=self.headers)
                if response.status_code == 200: break
        else:
            search_url = 'https://scholar.google.com/scholar'
            params = {
                'hl': 'en',
                'as_sdt': '0,5',
                'q': keyword,
                'btnG': ''
            }
            response = self.session.get(search_url, params=params, headers=self.headers)
        # parse
        soup = BeautifulSoup(response.text, features='lxml')
        papers = soup.find_all('div', class_='gs_r')
        paperinfos = []
        for paper in papers:
            try:
                pdf = paper.find('div', class_='gs_ggs gs_fl')
                link = paper.find('h3', class_='gs_rt')
                if link.find('a'): input_content = link.find('a')['href']
                elif pdf: input_content = pdf.find('a')['href']
                else: continue
                title = link.text
                authors = paper.find('div', class_='gs_a').text.split('\xa0')[0]
                paperinfo = {
                    'input': input_content,
                    'source': self.source,
                    'savedir': self.config['savedir'],
                    'ext': 'pdf',
                    'savename': title,
                    'title': title,
                    'authors': authors,
                    'download_url': None,
                }
            except:
                continue
            paperinfos.append(paperinfo)
            if len(paperinfos) == self.config['search_size_per_source']: break
        # return
        return paperinfos


'''
Function:
    Base class for the paper sources
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
import time
import requests
from ..utils import Downloader, colorize, Logger


'''Base class for the paper sources'''
class Base():
    def __init__(self, config, logger_handle, **kwargs):
        default_config = {
            'logfilepath': 'paperdl.log',
            'search_size_per_source': 5,
            'savedir': 'papers'
        }
        for key, value in default_config.items():
            if key not in config: config[key] = value
        self.source = None
        self.session = requests.Session()
        self.session.proxies.update(config['proxies'])
        self.config = config
        self.logger_handle = logger_handle
        if self.logger_handle is None: self.logger_handle = Logger(config['logfilepath'])
    '''search paper'''
    def search(self, keyword):
        raise NotImplementedError('not to be implemented')
    '''download paper'''
    def download(self, paperinfos):
        if hasattr(self, 'parseinfosbeforedownload'): paperinfos = self.parseinfosbeforedownload(paperinfos)
        default_config = {
            'savedir': 'papers',
            'savename': f'downloaded_{int(time.time())}',
            'ext': 'pdf',
            'source': self.source,
        }
        for paperinfo in paperinfos:
            for key, value in default_config.items():
                if key not in paperinfo: paperinfo[key] = value
            self.logger_handle.info(f"Downloading {colorize(paperinfo['savename'], 'highlight')} from {colorize(self.source.upper(), 'highlight')}")
            task = Downloader(paperinfo, self.session)
            if task.start():
                self.logger_handle.info(f"Downloaded {colorize(paperinfo['savename'], 'highlight')} from {colorize(self.source.upper(), 'highlight')} successfully")
            else:
                self.logger_handle.warning(f"Fail to download {colorize(paperinfo['savename'], 'highlight')} from {colorize(self.source.upper(), 'highlight')}")
    '''repr'''
    def __repr__(self):
        return 'Paper Source: %s' % self.source


'''
Function:
    Misc utils
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
import os
import re
import json


'''touch dir'''
def touchdir(dirpath):
    if not os.path.exists(dirpath):
        os.mkdir(dirpath)
        return False
    return True


'''load config file'''
def loadConfig(filepath='config.json'):
    f = open(filepath, 'r', encoding='utf-8')
    return json.load(f)


'''clear bad characters in filename'''
def filterBadCharacter(string):
    need_removed_strs = ['<em>', '</em>', '<', '>', '\\', '/', '?', ':', '"', '：', '|', '？', '*']
    for item in need_removed_strs:
        string = string.replace(item, '')
    try:
        rule = re.compile(u'[\U00010000-\U0010ffff]')
    except:
        rule = re.compile(u'[\uD800-\uDBFF][\uDC00-\uDFFF]')
    string = rule.sub('', string)
    return string.strip().encode('utf-8', 'ignore').decode('utf-8')


'''
Function:
    Paperdl: Search and download paper from specific websites
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
import sys
import copy
import json
import click
import warnings
import threading
if __name__ == '__main__':
    from modules import *
    from __init__ import __version__
else:
    from .modules import *
    from .__init__ import __version__
warnings.filterwarnings('ignore')


'''basic info'''
BASICINFO = '''************************************************************
Function: Paperdl V%s
Author: Charles
WeChat public account: Charles_pikachu
Help:
    input r: initialize the program
    input q: quit the program
    download: input IDs (e.g., '1, 2') to download multi papers
Savedir:
    The downloaded papers are saved in %s
************************************************************'''


'''Paperdl'''
class Paperdl():
    def __init__(self, configpath=None, config=None, **kwargs):
        assert configpath or config, 'configpath or config should be given'
        self.config = loadConfig(configpath) if config is None else config
        default_config = {
            'logfilepath': 'paperdl.log',
            'search_size_per_source': 5,
            'savedir': 'papers'
        }
        for key, value in default_config.items():
            if key not in self.config: self.config[key] = value
        self.logger_handle = Logger(self.config['logfilepath'])
        self.initializeAllSources()
    '''run'''
    def run(self, target_srcs=None):
        while True:
            print(BASICINFO % (__version__, self.config.get('savedir')))
            # search paper
            user_input = self.dealInput('Input the title of the paper: ')
            target_srcs = [
                'arxiv', 'googlescholar',
            ] if target_srcs is None else target_srcs
            search_results = self.search(user_input, target_srcs)
            # print search results
            title = ['ID', 'Title', 'First Author', 'Source']
            items, records, idx = [], {}, 0
            for key, values in search_results.items():
                for value in values:
                    items.append([
                        colorize(str(idx), 'number'), 
                        colorize(value['title'] if len(value['title']) < 50 else value['title'][:50] + '...', 'title'), 
                        value['authors'].split(',')[0], 
                        value['source'].upper()
                    ])
                    records.update({str(idx): value})
                    idx += 1
            if len(items) < 1: 
                self.logger_handle.error(colorize('No related papers were found', 'red'))
                continue
            printTable(title, items)
            # download papers
            user_input = self.dealInput('Input the paper ID you want to download: ')
            need_download_numbers = user_input.replace(' ', '').split(',')
            paperinfos = []
            for item in need_download_numbers:
                paperinfo = records.get(item, '')
                if paperinfo: paperinfos.append(paperinfo)
            self.download(paperinfos)
    '''search paper'''
    def search(self, keyword, target_srcs):
        self.logger_handle.info(f'Searching {colorize(keyword, "highlight")} From {colorize("|".join([c.upper() for c in target_srcs]), "highlight")}')
        def threadSearch(search_api, keyword, target_src, search_results):
            try:
                search_results.update({target_src: search_api(keyword)})
            except Exception as err:
                self.logger_handle.error(str(err), True)
                self.logger_handle.warning(f'Fail to search {colorize(keyword, "highlight")} from {colorize(target_src, "highlight")}')
        task_pool, search_results = [], {}
        for target_src in target_srcs:
            task = threading.Thread(
                target=threadSearch,
                args=(getattr(self, target_src).search, keyword, target_src, search_results)
            )
            task_pool.append(task)
            task.start()
        for task in task_pool:
            task.join()
        return search_results
    '''download paper'''
    def download(self, paperinfos):
        for paperinfo in paperinfos:
            getattr(self, paperinfo['source']).download([paperinfo])
    '''initialize all sources'''
    def initializeAllSources(self):
        supported_sources = {
            'arxiv': Arxiv,
            'scihub': SciHub,
            'googlescholar': GoogleScholar,
        }
        for key, value in supported_sources.items():
            setattr(self, key, value(copy.deepcopy(self.config), self.logger_handle))
        return supported_sources
    '''deal with user inputs'''
    def dealInput(self, tip=''):
        user_input = input(tip)
        if user_input.lower() == 'q':
            self.logger_handle.info('ByeBye')
            sys.exit()
        elif user_input.lower() == 'r':
            self.initializeAllSources()
            self.run()
        else:
            return user_input
    '''str'''
    def __str__(self):
        return 'Welcome to use paperdl!\nYou can visit https://github.com/CharlesPikachu/paperdl for more details.'


'''cmd直接运行'''
@click.command()
@click.version_option()
@click.option('-m', '--mode', default='download', help='the used mode, support "search" and "download"')
@click.option('-i', '--inp', default=None, help='the paper to download, the supported format is the same as sci-hub')
@click.option('-s', '--source', default=None, help='the used source, support "arxiv", "scihub" and "googlescholar", you can use "," to split multi sources')
@click.option('-d', '--savedir', default='papers', help='the directory for saving papers')
@click.option('-l', '--logfilepath', default='paperdl.log', help='the logging filepath')
@click.option('-z', '--size', default=5, help='search size per source')
@click.option('-p', '--proxies', default='{}', help='the proxies to be adopted')
@click.option('-a', '--area', default='CN', help='your area, support "CN" and "EN"')
def paperdlcmd(mode, inp, source, savedir, logfilepath, size, proxies, area):
    # prepare
    assert mode in ['search', 'download']
    area = area.upper()
    if mode == 'download': assert inp is not None, 'input url should be specified in download mode'
    config = {
        'logfilepath': logfilepath,
        'savedir': savedir,
        'search_size_per_source': size,
        'proxies': json.loads(proxies),
        'area': area,
    }
    if source is None: 
        target_srcs = ['arxiv', 'googlescholar']
    else:
        target_srcs = [s.strip() for s in source.split(',')]
    client = Paperdl(config=config)
    # if you select the search mode
    if mode == 'search':
        client.run(target_srcs=target_srcs)
    else:
        print(client)
        if source is None: source = 'scihub'
        paperinfo = {
            'savename': inp.strip('/').split('/')[-1],
            'ext': 'pdf',
            'savedir': savedir,
            'input': inp,
            'source': source,
        }
        client.download([paperinfo])


'''run'''
if __name__ == '__main__':
    import os
    rootdir = os.path.split(os.path.abspath(__file__))[0]
    client = Paperdl(os.path.join(rootdir, 'config.json'))
    client.run()


'''
Function:
    Some utils related with logging
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
import logging
from prettytable import PrettyTable


'''define the colors'''
COLORS = {
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'pink': '\033[35m',
    'cyan': '\033[36m',
    'highlight': '\033[93m',
    'number': '\033[96m',
    'title': '\033[93m',
}


'''Logger'''
class Logger():
    def __init__(self, logfilepath, **kwargs):
        setattr(self, 'logfilepath', logfilepath)
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s %(levelname)-8s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            handlers=[logging.FileHandler(logfilepath), logging.StreamHandler()],
        )
    @staticmethod
    def log(level, message):
        logging.log(level, message)
    def debug(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.DEBUG, message)
    def info(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.INFO, message)
    def warning(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.WARNING, message)
    def error(self, message, disable_print=False):
        if disable_print:
            fp = open(self.logfilepath, 'a')
            fp.write(message + '\n')
        else:
            Logger.log(logging.ERROR, message)


'''print table'''
def printTable(title, items):
    assert isinstance(title, list) and isinstance(items, list), 'title and items should be list'
    table = PrettyTable(title)
    for item in items: table.add_row(item)
    print(table)
    return table


'''colorize words in terminal'''
def colorize(string, color):
    string = str(string)
    if color not in COLORS: return string
    return COLORS[color] + string + '\033[0m'


'''
Function:
    Downloader
Author:
    Charles
WeChat public account:
    Charles_pikachu
'''
import os
import requests
from .misc import touchdir
from alive_progress import alive_bar


'''Downloader'''
class Downloader():
    def __init__(self, paperinfo, session=None, **kwargs):
        self.paperinfo = paperinfo
        self.session = requests.Session() if session is None else session
        self.__setheaders(paperinfo['source'])
    '''start to download according to paperinfo'''
    def start(self):
        paperinfo, session, headers = self.paperinfo, self.session, self.headers
        if not paperinfo['download_url']: return False
        touchdir(paperinfo['savedir'])
        with session.get(paperinfo['download_url'], headers=headers, stream=True) as response:
            if response.status_code not in [200]: return False
            total_size, chunk_size, downloaded_size = int(response.headers['content-length']), paperinfo.get('chunk_size', 1024), 0
            savepath = os.path.join(paperinfo['savedir'], f"{paperinfo['savename']}.{paperinfo['ext']}")
            text, fp = '[FileSize]: %0.2fMB/%0.2fMB', open(savepath, 'wb')
            with alive_bar(manual=True) as bar:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if not chunk: continue
                    fp.write(chunk)
                    downloaded_size += chunk_size
                    bar.text(text % (downloaded_size / 1024 / 1024, total_size / 1024 / 1024))
                    bar(min(downloaded_size / total_size, 1))
        return True
    '''set request headers'''
    def __setheaders(self, source):
        if hasattr(self, f'{source}_headers'):
            self.headers = getattr(self, f'{source}_headers')
        else:
            self.headers = {
                'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36'
            }


'''
Function:
    setup the paperdl
Author:
    Charles
微信公众号:
    Charles的皮卡丘
GitHub:
    https://github.com/CharlesPikachu/paperdl
'''
import paperdl
from setuptools import setup, find_packages


'''readme'''
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()


'''setup'''
setup(
    name=paperdl.__title__,
    version=paperdl.__version__,
    description=paperdl.__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent'
    ],
    author=paperdl.__author__,
    url=paperdl.__url__,
    author_email=paperdl.__email__,
    license=paperdl.__license__,
    include_package_data=True,
    entry_points={'console_scripts': ['paperdl = paperdl.paperdl:paperdlcmd']},
    install_requires=list(open('requirements.txt', 'r').readlines()),
    zip_safe=True,
    packages=find_packages()
)